# Macro precision, recall, f-measure - calculated separated
# Calculatins separated for each evaluation window
# Without considering novel classes
# -------------------------------------------------------------------------------------
macro.precision.recall.fmeasure.windows <- function(true,predicted,num.evaluation.windows){
  
  num.labels <- ncol(true)
  num.examples <- nrow(true)
  result <- list()
  ma.precision.window <- vector()
  ma.recall.window <- vector()
  ma.fmeasure.window <- vector()
  
  # Number of examples per window. Try to keep an equal number of examples per window
  num.examples.window <- floor(num.examples / num.evaluation.windows)
  evaluation.windows <- rep(num.examples.window,num.evaluation.windows)
  rest <- nrow(predicted) - (num.examples.window*num.evaluation.windows)
  if(rest > 0){
    for(i in 1:rest){
      evaluation.windows[i] <- evaluation.windows[i]+1
    }
  }
  
  ini <- 1
  end <- 0
  tp <- vector(length=num.labels, mode="numeric")
  fp <- vector(length=num.labels, mode="numeric")
  fn <- vector(length=num.labels, mode="numeric")
  beta <- 1
  
  for(i in 1:num.evaluation.windows){
    end <- end + evaluation.windows[i]
    
    predictions.window <- predicted[ini:end,,drop=F]
    true.classes.window <- true[ini:end,,drop=F]
    num.examples.local <- nrow(predictions.window)
    
    total.prec <- 0
    total.reca <- 0
    total.fmeas <- 0
    
    for(j in 1:num.labels){
      
      prec <- 0
      reca <- 0
      fmeas <- 0
      
      # Index the number of each example evaluated
      index.example <- ini
      
      for(k in 1:num.examples.local){
        
        index.example <- index.example + 1
        
        if(true.classes.window[k,j]==1 && predictions.window[k,j]==1){
          tp[j] <- tp[j]+1
        }
        else if(true.classes.window[k,j]==0 && predictions.window[k,j]==1){
          fp[j] <- fp[j]+1
        }
        else if(true.classes.window[k,j]==1 && predictions.window[k,j]==0){
          fn[j] <- fn[j]+1
        }
      }
      
      # Treat special cases just like Mulan
      # (i) if the algorithm outputs the empty set and the ground truth is the empty set,
      # then we consider precision equal to 1 (ii) if the algorithm outputs the
      # empty set and the ground truth is not empty, then we consider precision
      # equal to 0.
      if (tp[j] + fp[j] + fn[j] == 0) {
        prec <- 1
      }
      else if (tp[j] + fp[j] == 0) {
        prec <- 0
      }
      else{
        prec <- tp[j] / (tp[j] + fp[j])
      }
      
      # Treat special cases just like Mulan
      # (i) if the algorithm outputs the empty set and the ground truth is the empty set,
      # then we consider recall equal to 1 (ii) if the algorithm does not output
      # the empty set and the ground truth is empty, then we consider recall
      # equal to 0.  
      if (tp[j] + fp[j] + fn[j] == 0) {
        reca <- 1
      }
      else if (tp[j] + fn[j] == 0) {
        reca <- 0
      }
      else{
        reca <- tp[j] / (tp[j] + fn[j])
      } 
      
      # Treat special cases just like Mulan
      # (i) if the algorithm outputs the empty set and the ground truth is the empty
      # set, then we consider F equal to 1 (ii) if the algorithm outputs the
      # empty set and the ground truth is not empty, then we consider F equal to
      # 0 (iii) if the ground truth is empty and the algorithm does not output
      # the empty set, then we consider F equal to 0 (iv) if neither the ground
      # truth nor the algorithm's prediction is the empty set and their
      # intersection is empty, then we consider F equal to 0.
      if (tp[j] + fp[j] + fn[j] == 0) {
        fmeas <- 1
      }
      else{
        beta2 = beta * beta
        fmeas <- ((beta2 + 1) * tp[j]) / ((beta2 + 1) * tp[j] + beta2 * fn[j] + fp[j])
      }
      
      total.prec <- total.prec + prec
      total.reca <- total.reca + reca
      total.fmeas <- total.fmeas + fmeas
    }
    
    precision <- total.prec / num.labels
    recall <- total.reca / num.labels
    fmeasure <- total.fmeas / num.labels
    
    ma.precision.window <- c(ma.precision.window,precision)
    ma.recall.window <- c(ma.recall.window,recall)
    ma.fmeasure.window <- c(ma.fmeasure.window,fmeasure)
    
    ini <- end + 1
  }
  
  result$ma.precision <- total.prec / num.labels
  result$ma.recall <- total.reca / num.labels
  result$ma.fmeasure <- total.fmeas / num.labels
  
  result$ma.precision.window <- ma.precision.window
  result$ma.recall.window <- ma.recall.window
  result$ma.fmeasure.window <- ma.fmeasure.window
  
  return(result)	
  
}
# -------------------------------------------------------------------------------------


# Update model information when an example is explained
# -------------------------------------------------------------------------------------
update.model.information <- function(mapping,pred,x,time.stamp,n0,initial.number.classes,predictions,winner,inst.l){
  
  # Get neurons and distances
  neurons <- as.numeric(winner$nn.index[inst.l,])
  distances <- as.numeric(winner$nn.dist[inst.l,])
  
  # For each neuron
  for(idx in 1:length(neurons)){
    
    # Index neuron
    idx.neuron <- neurons[idx]
    
    # Distance from neuron
    distance <- distances[idx]
    
    # Micro clusters
    mapping$som.micro.clusters[[1]][[2]][[idx.neuron]][[3]] <- mapping$som.micro.clusters[[1]][[2]][[idx.neuron]][[3]] + 1
    mapping$som.micro.clusters[[1]][[2]][[idx.neuron]][[4]] <- time.stamp
    
    # Udpate according to distance
    delta <- n0 * (x - mapping$som.micro.clusters[[1]][[1]]$codes[[1]][idx.neuron,]) * exp(-distance)
    old <- mapping$som.micro.clusters[[1]][[1]]$codes[[1]][idx.neuron,]
    mapping$som.micro.clusters[[1]][[1]]$codes[[1]][idx.neuron,] <- old + delta
    mapping$som.micro.clusters[[1]][[2]][[idx.neuron]][[1]] <- old + delta
    
  }
  
  return(mapping)
}
# -------------------------------------------------------------------------------------



# Update class totals and class probabilities matrices
# -------------------------------------------------------------------------------------
updateClassTotalsProbabilities <- function(mapping,pred,num.pred,predictions,initial.number.classes,is.novelty,num.offline.instances){
  
  # Total number of instances explained
  mapping$total.instances <- mapping$total.instances + num.pred
  
  # If there is a novelty, we already added the examples to mapping$total.instances.NP
  if(is.novelty == 0){
    mapping$total.instances.NP <- mapping$total.instances.NP + num.pred
  }
  
  # Total number of examples predicted
  num.predictions <- mapping$total.instances - num.offline.instances
  
  if(nrow(pred) > 0){
    for(r in 1:nrow(pred)){
      # Neurons which explained the example
      idx.classes <- which(pred[r,] > 0)
      
      if(length(idx.classes) > 0){
        # Class totals
        for(i in 1:length(idx.classes)){
          for(j in 1:length(idx.classes)){
            idx.i <- idx.classes[i]
            idx.j <- idx.classes[j]
            if( pred[r,idx.i] == 1 | pred[r,idx.j] == 1 ){ 
              mapping$class.totals[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] + 1
              # Otherwise we have already counted Totals[c_i,c_i]
            }
          }
        }
      }
    }
  }
  
  # We need to check if we are dealing with an existing class or with a NP
  # The probabilities are calculated differently
  for(idx.i in 1:nrow(mapping$class.totals)){
    for(idx.j in 1:nrow(mapping$class.totals)){
      
      if(idx.i > initial.number.classes & idx.j < initial.number.classes){ # dealing with a NP idx.i
        mapping$class.probabilities[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] / mapping$class.totals[idx.j,idx.j]
      }
      else if(idx.i > initial.number.classes & idx.j > initial.number.classes){ # dealing with two NP idx.i and idx.j
        if(idx.i == idx.j){
          # Predictions so far. We have to consider only the predictions from where NP appeared until the current time stamp
          mapping$class.probabilities[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] / mapping$total.instances.NP[idx.j-initial.number.classes]
        }
        else{
          mapping$class.probabilities[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] / mapping$class.totals[idx.j,idx.j]
        }
      }
      else if(idx.i < initial.number.classes & idx.j > initial.number.classes){ # dealing with a NP idx.j
        mapping$class.probabilities[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] / mapping$class.totals[idx.j,idx.j]
      }
      else{ # Dealing with existing classes
        if(idx.i == idx.j){
          mapping$class.probabilities[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] / mapping$total.instances
        }
        else{
          mapping$class.probabilities[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] / mapping$class.totals[idx.j,idx.j]
        }
      }
    }
  }
  
  return(mapping)
}



# Get parameter values from file
# -------------------------------------------------------------------------------------
getParameterValues <- function(param.file){
  
  parameters <- c("train.mode","n0","n1","omega","num.iterations","theta","min.examples.cluster",
                  "train.data","test.data","num.classes","novel.classes","num.evaluation.windows","window.stm.check",
                  "num.experiments","empty.predictions","grid.dimension","update.model.info")
  
  values <- list()
  
  train.mode <- 0
  num.classes <- 0
  novel.classes <- 0
  n0 <- 0
  n1 <- 0
  grid.dimension <- 0
  num.iterations <- 0
  omega <- 0
  theta <- 0
  min.examples.cluster <- 0
  train.data <- 0
  test.data <- 0
  num.evaluation.windows <- 0
  window.stm.check <- 0
  num.experiments <- 0
  empty.predictions <- 0
  update.model.info <- 0
  
  # parameter file
  content <- scan(param.file,what="character",sep="\n")
  
  for(i in 1:length(parameters)){
    
    idx <- grep(parameters[i],content)
    value <- unlist(strsplit(content[idx],"="))[2]
    value <- gsub(" ","",value)
    
    if(parameters[i] == "n0"){
      n0 <- as.numeric(value)
    }
    else if(parameters[i] == "num.iterations"){
      num.iterations <- as.numeric(value)
    }
    else if(parameters[i] == "num.experiments"){
      num.experiments <- as.numeric(value)
    }
    else if(parameters[i] == "num.evaluation.windows"){
      num.evaluation.windows <- as.numeric(value)
    }
    else if(parameters[i] == "window.stm.check"){
      window.stm.check <- as.numeric(value)
    }
    else if(parameters[i] == "n1"){
      n1 <- as.numeric(value)
    }
    else if(parameters[i] == "train.data"){
      train.data <- value
    }
    else if(parameters[i] == "test.data"){
      test.data <- value
    }
    else if(parameters[i] == "train.mode"){
      train.mode <- value
    }
    else if(parameters[i] == "omega"){
      omega <- as.numeric(value)
    }
    else if(parameters[i] == "theta"){
      theta <- as.numeric(value)
    }
    else if(parameters[i] == "radius.factor"){
      radius.factor <- as.numeric(value)
    }
    else if(parameters[i] == "nd.radius.factor"){
      nd.radius.factor <- as.numeric(value)
    }
    else if(parameters[i] == "win.neuron.threshold"){
      win.neuron.threshold <- as.numeric(value)
    }
    else if(parameters[i] == "conditional.prob.threshold"){
      conditional.prob.threshold <- as.numeric(value)
    }
    else if(parameters[i] == "min.examples.cluster"){
      min.examples.cluster <- as.numeric(value)
    }
    else if(parameters[i] == "num.classes"){
      num.classes <- as.numeric(value)
    }
    else if(parameters[i] == "novel.classes"){
      novel.classes <- as.numeric(unlist(strsplit(value,",")))
    }
    else if(parameters[i] == "grid.dimension"){
      grid.dimension <- as.numeric(value)
    }
    else if(parameters[i] == "empty.predictions"){
      empty.predictions <- as.numeric(value)
    }
    else if(parameters[i] == "update.model.info"){
      update.model.info <- as.numeric(value)
    }
  }  
  
  values$n0 <- n0
  values$n1 <- n1
  values$num.iterations <- num.iterations
  values$omega <- omega
  values$theta <- theta
  values$min.examples.cluster <- min.examples.cluster
  values$train.data <- train.data
  values$test.data <- test.data
  values$num.classes <- num.classes
  values$novel.classes <- novel.classes
  values$train.mode <- train.mode
  values$num.evaluation.windows <- num.evaluation.windows
  values$window.stm.check <- window.stm.check
  values$num.experiments <- num.experiments
  values$empty.predictions <- empty.predictions
  values$grid.dimension <- grid.dimension
  values$update.model.info <- update.model.info
  
  return(values)
}

# Compute micro-cluster properties
# -------------------------------------------------------------------------------------
computeMicroClusters <- function(som.map,offline.classes,min.ex){
  # Will store properties for each micro cluster
  # Each class will be represented by a set of neurons, whose mapped examples form micro-clusters
  # Properties: 1-LS, 2-SS, 3-n, 4-t, 5-prototype vector representing classes, 6-Radius, 7-time.stamp.creation, 8-class position in the matrices with properties,
  #             9-radius factor for novelty detection 1, 10-conditional probability threshold for each class, 11-average output [sum outputs, num examples], 
  #             12-radius factor for novelty detection 2
  
  
  # Get neurons to remove
  table.som.map.classif <- table(som.map$unit.classif)
  neurons <- as.numeric(names(table.som.map.classif[table.som.map.classif >= min.ex]))
  idx.neuros.to.remove <- as.numeric(names(table.som.map.classif[table.som.map.classif < min.ex]))
  
  # Mark neurons not used
  for(i in idx.neuros.to.remove){
    som.map$unit.classif[som.map$unit.classif==i] <- 0
  }
  
  # Structure for micro clusters
  micro.clusters <- lapply(1:length(neurons), function(x) as.list(rep(0,12)))
  
  # Store results
  results <- list()
  
  # There may be empty neurons! Lets keep only the used ones
  som.map$codes[[1]] <- som.map$codes[[1]][neurons,,drop=F] 
  
  # Go through all neurons
  pos <- 0
  for(idx in neurons){
    indexes <- which(som.map$unit.classif==idx)
    pos <- pos + 1
    micro.clusters[[pos]][[1]] <- as.numeric(som.map$codes[[1]][pos,])
    
    micro.clusters[[pos]][[3]] <- length(indexes) # number of instances
    micro.clusters[[pos]][[5]] <- as.numeric(apply(offline.classes[indexes,],2,mean)) # prototype vector representing classes
    micro.clusters[[pos]][[7]] <- 0 # Time stamp when created (Will not be modified)
    micro.clusters[[pos]][[8]] <- 0 # class position in the matrices with properties (not used)
    micro.clusters[[pos]][[10]] <- rep(9,ncol(offline.classes))# conditional probability threshold for each class  
    
    # We need to do that if we removed codebooks
    som.map$unit.classif[which(som.map$unit.classif==idx)] <- pos
  }
  
  results$som.map <- som.map
  results$micro.clusters <- micro.clusters
  
  return(results)
}
# -------------------------------------------------------------------------------------


# For each neuron and each class, updates its conditional probabilitiy thresholds
# We will consider a p(x|y_j) = average.output.som.map for all neurons
# That means we are considering the output of the neurons to be average.output.som.map to calculate the threshold
# We consider that all classes are predicted to calculate the thresholds
# p(y_j | y_k, x) will be weighted by the corresponding value for y_j in the prototype vector of the neuron
# -------------------------------------------------------------------------------------------------
updateCondProbabilitiesNeurons <- function(micro.clusters,class.probabilities){
  
  for(i in 1:length(micro.clusters)){
    
    prototype.vector <- micro.clusters[[i]][[5]]
    idx.classes.mapped <- which(prototype.vector > 0)
    
    for(idx.class in idx.classes.mapped){
      
      # Return the probabilities sorted
      class.prob <- class.probabilities[idx.class,]
      class.prob[idx.class] <- 0
      prob.sorted <- sort(class.prob,decreasing=TRUE,index.return=TRUE)
      idx.j <- idx.class
      prob.j <- class.probabilities[idx.j,idx.j] # p(y_j)
      
      prob.k.j <- 1
      for(k in idx.classes.mapped){ # p(y_k|y_j)
        idx.k <- prob.sorted$ix[k]
        if(class.probabilities[idx.k,idx.j] > 0 & idx.k != idx.j){
          prob.k.j <- prob.k.j * class.probabilities[idx.k,idx.j]
        }  
      }
      
      weight.factor <- micro.clusters[[i]][[5]][idx.class] # prototype vector
      prob.x.j <- micro.clusters[[i]][[11]][1] / micro.clusters[[i]][[11]][2] # p(x|y_j)
      prob.j.ks.x <- prob.j * prob.k.j * prob.x.j #p(y_j | y_k, x)
      
      #micro.clusters[[i]][[10]][idx.class] <- prob.j.ks.x * weight.factor
      micro.clusters[[i]][[10]][idx.class] <- prob.j.ks.x * exp(-(1-weight.factor))
      #micro.clusters[[i]][[10]][idx.class] <- prob.j.ks.x * exp(-weight.factor) # Não faz sentido
    }
  }
  
  return(micro.clusters)
  
}
# -------------------------------------------------------------------------------------------------


# For each neuron calculates its conditional probabilitiy thresholds for a given class
# We will consider a p(x|y_j) = average.output.som.map for all neurons
# That means we are considering the output of the neurons to be average.output.som.map to calculate the threshold
# We consider that all classes are predicted to calculate the thresholds
# p(y_j | y_k, x) will be weighted by the corresponding value for y_j in the prototype vector of the neuron
# -------------------------------------------------------------------------------------------------
getCondProbabilitiesNeurons <- function(micro.clusters,class.probabilities,average.output.som.map){
  
  for(i in 1:length(micro.clusters)){
    
    prototype.vector <- micro.clusters[[i]][[5]]
    idx.classes.mapped <- which(prototype.vector > 0)
    
    for(idx.class in idx.classes.mapped){
      
      # Return the probabilities sorted
      class.prob <- class.probabilities[idx.class,]
      class.prob[idx.class] <- 0
      prob.sorted <- sort(class.prob,decreasing=TRUE,index.return=TRUE)
      idx.j <- idx.class
      prob.j <- class.probabilities[idx.j,idx.j] # p(y_j)
      
      prob.k.j <- 1
      for(k in idx.classes.mapped){ # p(y_k|y_j)
        idx.k <- prob.sorted$ix[k]
        if(class.probabilities[idx.k,idx.j] > 0 & idx.k != idx.j){
          prob.k.j <- prob.k.j * class.probabilities[idx.k,idx.j]
        }  
      }
      
      weight.factor <- micro.clusters[[i]][[5]][idx.class]
      prob.x.j <- average.output.som.map[[i]][1] / average.output.som.map[[i]][2] # p(x|y_j)
      prob.j.ks.x <- prob.j * prob.k.j * prob.x.j #p(y_j | y_k, x)
      
      #micro.clusters[[i]][[10]][idx.class] <- prob.j.ks.x * weight.factor
      micro.clusters[[i]][[10]][idx.class] <- prob.j.ks.x * exp(-(1-weight.factor))
      #micro.clusters[[i]][[10]][idx.class] <- prob.j.ks.x * exp(-weight.factor) # não faz sentido
      micro.clusters[[i]][[11]] <- average.output.som.map[[i]]
    }
  }
  
  return(micro.clusters)
  
  
  
#   z <- nrow(class.probabilities)
#   #z <- ceiling(z)
#   reduction.factor <- 1
#   
#   # For each class
#   for(idx.class in 1:ncol(class.probabilities)){
#     
#     # Return the probabilities sorted
#     class.prob <- class.probabilities[idx.class,]
#     class.prob[idx.class] <- 0
#     prob.sorted <- sort(class.prob,decreasing=TRUE,index.return=TRUE)
#     idx.j <- idx.class
#     prob.j <- class.probabilities[idx.j,idx.j] # p(y_j)
#     
#     prob.k.j <- 1
#     for(k in 1:z){ # p(y_k|y_j)
#       idx.k <- prob.sorted$ix[k]
#       if(class.probabilities[idx.k,idx.j] > 0 & idx.k != idx.j){
#         prob.k.j <- prob.k.j * class.probabilities[idx.k,idx.j]
#       }  
#     }
#     
#     # Go through all micro-clusters
#     for(i in 1:length(micro.clusters)){
#       weight.factor <- micro.clusters[[i]][[5]][idx.class]
#       prob.x.j <- average.output.som.map[i] # p(x|y_j)
#       prob.j.ks.x <- prob.j * prob.k.j * prob.x.j #p(y_j | y_k, x)
#       
#       if(weight.factor > 0){ # There were mapped instances classified in class idx.class
#         micro.clusters[[i]][[10]][idx.class] <- prob.j.ks.x * reduction.factor * (1-weight.factor)
#       }
#       else{ # There were no mapped instances classified in class idx.class
#         micro.clusters[[i]][[10]][idx.class] <- 99 # New instances if mapped to this neuron cannot be classified in class idx.class
#       }
#     }
#   }
#  return(micro.clusters)
}
# -------------------------------------------------------------------------------------------------


# For each neuron, compute its average output
# -------------------------------------------------------------------------------------------------
getAverageNeuronOutputs <- function(output.som.map,num.micro.clusters){
  
  # Store the average outputs
  #average.output.som.map <- vector(length=num.micro.clusters,mode="numeric")
  average.output.som.map <- vector("list", num.micro.clusters)
  
  for(i in 1:num.micro.clusters){
    distances <- output.som.map$distances[as.numeric(which(output.som.map$unit.classif==i))]
    average.output.som.map[[i]] <- c(sum(sapply(distances,function(x){return(exp(-x))})), length(distances))
      #mean(sapply(distances,function(x){return(exp(-x))}))
  }
  
  return(average.output.som.map)
}
# -------------------------------------------------------------------------------------------------


# Compute probabilities p(i) and p(i|j)
# Also counts total of examples per classes and intersections
# -------------------------------------------------------------------------------------
# classes data frame, index of classes
get.probabilities <- function(classes,i,j){
  
  result <- list()
  result$prob <- 0
  result$total <- 0
  
  if(i == j){ # p(i)
    total.i <- length(which(classes[,i]==1))
    total <- nrow(classes)
    
    result$total <- total.i
    result$prob <- total.i / total
  }
  else{ # p(i|j)
    total.j <- length(which(classes[,j]==1))
    examples.i <- which(classes[,i]==1)
    examples.j <- which(classes[,j]==1)
    intersect.i.j <- intersect(examples.i,examples.j)
    total.i.j <- length(intersect.i.j)
    
    result$total <- total.i.j
    result$prob <- total.i.j / total.j
  }	
  
  return(result)
}
# -------------------------------------------------------------------------------------

# Compute label cardinality
# -------------------------------------------------------------------------------------
computeLabelCardinality <- function(offline.classes){
  
  z <- 0
  
  for(i in 1:nrow(offline.classes)){
    labels <- offline.classes[i,]
    idx.classified <- which(labels == 1)
    z <- z + length(idx.classified)
  }
  
  z <- z / nrow(offline.classes)
  
  return(z)
}

# -------------------------------------------------------------------------------------

# Compute initial class totals and class probabilities
# -------------------------------------------------------------------------------------------------
computeInitialClassProbabilitiesTotals <- function(offline.classes){
  
  result <- list()
  num.classes <- ncol(offline.classes) # number of initial classes. 
  
  # Fill class probabilities matrix
  # Also fill the total counts of examples per class and intersections.
  # This information will be used to further update class probabilities
  class.probabilities <- matrix(nrow=num.classes, ncol=num.classes, data=0)
  class.totals <- matrix(nrow=num.classes, ncol=num.classes, data=0)
  for(i in 1:num.classes){
    for(j in 1:num.classes){
      class.prob.totals <- get.probabilities(offline.classes,i,j)  # p(i) or p(i|j) and totals
      class.probabilities[i,j] <- class.prob.totals$prob
      class.totals[i,j] <- class.prob.totals$total
    }
  }
  
  # Name class totals and class probabilities columns and rows
  rownames(class.probabilities) <- paste("C",c(1:ncol(offline.classes)),sep="")
  colnames(class.probabilities) <- paste("C",c(1:ncol(offline.classes)),sep="")
  rownames(class.totals) <- paste("C",c(1:ncol(offline.classes)),sep="")
  colnames(class.totals) <- paste("C",c(1:ncol(offline.classes)),sep="")
  
  result$class.totals <- class.totals
  result$class.probabilities <- class.probabilities
  
  return(result)
}
# -------------------------------------------------------------------------------------------------


# Euclidean distance
# -------------------------------------------------------------------------------------
euc.dist <- function(x, y){
  
  distance <- sqrt(sum((x - y)^2))
  
  return(distance)
}
# -------------------------------------------------------------------------------------

# Computes radius factors for each neuron, based on the training examples these neurons classified
# -------------------------------------------------------------------------------------------------
computeRadiusFactorMC <- function(micro.clusters,output.som.map,average.output.som.map,data){
  
  for(i in 1:length(micro.clusters)){
    
    # Distances of codebook vectors to classified mapped instances
    distances <- output.som.map$distances[as.numeric(which(output.som.map$unit.classif==i))]
    
    # Indexes of examples mapped to neuron i
    indexes.data.mapped <- which(output.som.map$unit.classif==i)
    
    # Get max distance that classifies the farthest instance
    rFact <- max(distances)
    
    # Search for the most isolated point
    data.mapped <- data[indexes.data.mapped,,drop=F]
    
    # Centroid of the micro-cluster (weight vector)
    centroid <- micro.clusters[[i]][[1]]
    
    if(length(indexes.data.mapped) <= 1){browser()}
    
    # Get the nearest neighbour of each point
    d2 <- get.knn(data.mapped,k=1)
    
    # Search for isolated and its corresponding farthest neighbour
    isolated.index <- which(d2$nn.dist[,1]==max(d2$nn.dist[,1]))
    
    # If more than one, the isolated one will be the farthest from the centroid
    if(length(isolated.index) > 1){
      d3 <- as.matrix(dist(rbind(data.mapped[isolated.index,],centroid)))[length(isolated.index)+1,]
      isolated.index <- isolated.index[which(d3==max(d3))]
    }  
    
    # Position of the farthest.index index
    pos <- isolated.index
    
    # Search for nd.rFact 
    nd.rFact <- 9999999999999
    d2.sorted <- sort(d2$nn.dist[,1],decreasing=TRUE,index.return=TRUE)
    previous <- 0
    next.index <- 2
    checking <- 0
    
    while(nd.rFact >= rFact){
      checking <- checking + 1
      
      # Corresponding farthest neighbour
      farthest.index <- d2$nn.index[pos,]
      if(farthest.index == previous){ # Oh! We are in circles!
        next.index <- next.index + 1
        farthest.index <- d2.sorted$ix[next.index]
      }
      
      # Get distance between isolated.index and micro-cluster centroid
      nd.rFact <- euc.dist(data.mapped[farthest.index,],centroid)
      if(is.na(nd.rFact)){browser()}
      
      previous <- pos
      pos <- farthest.index
      
      if(checking==100){browser()}
    }
    
    if(nd.rFact == rFact){browser()}
    
    # Store radius factor and average output
    micro.clusters[[i]][[9]] <- rFact
    #micro.clusters[[i]][[11]] <- average.output.som.map[[i]]
    micro.clusters[[i]][[12]] <- nd.rFact 
    
    #plot(data.mapped)
    #points(centroid[1],centroid[2],col=2)
    #points(data.mapped[isolated.index,1],data.mapped[isolated.index,2],col=3)
    #points(data.mapped[farthest.index,1],data.mapped[farthest.index,2],col=4)
    #browser()
    
    if(rFact < nd.rFact){browser()}
  }
  
  return(micro.clusters)
}
# -------------------------------------------------------------------------------------------------